import datetime
from io import BytesIO
from typing import Optional

import pytest
import tzlocal
from asn1crypto import algos, cms, keys, ocsp
from asn1crypto.pdf import RevocationInfoArchival
from certomancer.integrations.illusionist import Illusionist
from certomancer.registry import (
    ArchLabel,
    CertLabel,
    EntityLabel,
    KeyLabel,
    ServiceLabel,
)
from freezegun import freeze_time
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
from pyhanko.pdf_utils.reader import PdfFileReader
from pyhanko.sign import PdfTimeStamper, signers, timestamps
from pyhanko.sign.ades.api import CAdESSignedAttrSpec
from pyhanko.sign.ades.report import (
    AdESFailure,
    AdESIndeterminate,
    AdESPassed,
    AdESStatus,
)
from pyhanko.sign.signers.pdf_cms import (
    GenericPdfSignedAttributeProviderSpec,
    PdfCMSSignedAttributes,
    SimpleSigner,
)
from pyhanko.sign.validation import SignatureCoverageLevel, ades
from pyhanko.sign.validation.policy_decl import (
    LocalKnowledge,
    PdfSignatureValidationSpec,
    SignatureValidationSpec,
)

from pyhanko_certvalidator import policy_decl as certv_policy_decl
from pyhanko_certvalidator.authority import CertTrustAnchor
from pyhanko_certvalidator.context import (
    CertValidationPolicySpec,
    ValidationContext,
)
from pyhanko_certvalidator.fetchers.requests_fetchers import (
    RequestsFetcherBackend,
)
from pyhanko_certvalidator.ltv.poe import (
    KnownPOE,
    POEType,
    ValidationObject,
    ValidationObjectType,
    digest_for_poe,
)
from pyhanko_certvalidator.path import ValidationPath
from pyhanko_certvalidator.policy_decl import (
    AlgorithmUsageConstraint,
    AlgorithmUsagePolicy,
    FreshnessReqType,
)
from pyhanko_certvalidator.registry import (
    SimpleCertificateStore,
    SimpleTrustManager,
)
from pyhanko_certvalidator.validate import async_validate_path

from .samples import (
    CERTOMANCER,
    MINIMAL_ONE_FIELD,
    TESTING_CA,
    UNRELATED_TSA,
)
from .signing_commons import (
    DUMMY_HTTP_TS_VARIANT,
    DUMMY_TS,
    DUMMY_TS2,
    FROM_CA,
    INTERM_CERT,
    REVOKED_SIGNER,
    TRUST_ROOTS,
    TSA_CERT,
    live_testing_vc,
)
from .test_pades import PADES


async def _generate_pades_test_doc(requests_mock, signer=FROM_CA, **kwargs):
    kwargs.setdefault('use_pades_lta', True)
    kwargs.setdefault('embed_validation_info', True)
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    vc = live_testing_vc(requests_mock)

    timestamper = timestamps.DummyTimeStamper(
        tsa_cert=TSA_CERT,
        tsa_key=TESTING_CA.key_set.get_private_key('tsa'),
        certs_to_embed=FROM_CA.cert_registry,
        override_md='sha256',
    )
    return await signers.async_sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            validation_context=vc,
            subfilter=PADES,
            **kwargs,
        ),
        signer=signer,
        timestamper=timestamper,
    )


async def _update_pades_test_doc(requests_mock, out):
    r = PdfFileReader(out)

    vc = live_testing_vc(requests_mock)
    await PdfTimeStamper(DUMMY_TS2).async_update_archival_timestamp_chain(r, vc)


DEFAULT_REVINFO_POLICY = certv_policy_decl.CertRevTrustPolicy(
    revocation_checking_policy=certv_policy_decl.REQUIRE_REVINFO,
    freshness_req_type=FreshnessReqType.MAX_DIFF_REVOCATION_VALIDATION,
)
DEFAULT_SIG_VALIDATION_SPEC = SignatureValidationSpec(
    cert_validation_policy=CertValidationPolicySpec(
        trust_manager=SimpleTrustManager.build(TRUST_ROOTS),
        revinfo_policy=DEFAULT_REVINFO_POLICY,
    )
)
DEFAULT_PDF_VALIDATION_SPEC = PdfSignatureValidationSpec(
    signature_validation_spec=DEFAULT_SIG_VALIDATION_SPEC
)


@pytest.mark.asyncio
async def test_pades_basic_happy_path(requests_mock):
    with freeze_time('2020-11-20'):
        w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
        out = await signers.async_sign_pdf(
            w,
            signers.PdfSignatureMetadata(field_name='Sig1', subfilter=PADES),
            signer=FROM_CA,
            timestamper=DUMMY_TS,
        )

    with freeze_time('2020-11-25'):
        r = PdfFileReader(out)
        live_testing_vc(requests_mock)
        result = await ades.ades_basic_validation(
            r.embedded_signatures[0].signed_data,
            validation_spec=DEFAULT_SIG_VALIDATION_SPEC,
            raw_digest=r.embedded_signatures[0].compute_digest(),
        )
        assert result.ades_subindic == AdESPassed.OK


@pytest.mark.asyncio
async def test_embedded_cades_happy_path(requests_mock):
    with freeze_time('2020-11-01'):
        signature = await FROM_CA.async_sign_general_data(
            b'Hello world!', 'sha256', detached=False, use_cades=True
        )
        signature = cms.ContentInfo.load(signature.dump())

    with freeze_time('2020-11-25'):
        live_testing_vc(requests_mock)
        result = await ades.ades_basic_validation(
            signature['content'],
            validation_spec=DEFAULT_SIG_VALIDATION_SPEC,
        )
        assert result.ades_subindic == AdESPassed.OK


@pytest.mark.asyncio
async def test_embedded_cades_with_time_happy_path(requests_mock):
    with freeze_time('2020-11-01'):
        signature = await FROM_CA.async_sign_general_data(
            b'Hello world!',
            'sha256',
            detached=False,
            timestamper=DUMMY_TS,
            use_cades=True,
        )
        signature = cms.ContentInfo.load(signature.dump())

    with freeze_time('2020-11-25'):
        live_testing_vc(requests_mock)
        result = await ades.ades_with_time_validation(
            signature['content'],
            validation_spec=DEFAULT_SIG_VALIDATION_SPEC,
        )
        assert result.ades_subindic == AdESPassed.OK
        assert result.best_signature_time == datetime.datetime(
            2020, 11, 1, tzinfo=datetime.timezone.utc
        )


@pytest.mark.asyncio
async def test_embedded_cades_provably_revoked(requests_mock):
    with freeze_time('2020-12-10'):
        signature = await REVOKED_SIGNER.async_sign_general_data(
            b'Hello world!',
            'sha256',
            detached=False,
            timestamper=DUMMY_TS,
            use_cades=True,
            signed_attr_settings=PdfCMSSignedAttributes(
                cades_signed_attrs=CAdESSignedAttrSpec(timestamp_content=True)
            ),
        )
        signature = cms.ContentInfo.load(signature.dump())

    with freeze_time('2020-12-25'):
        live_testing_vc(requests_mock)
        result = await ades.ades_basic_validation(
            signature['content'],
            validation_spec=DEFAULT_SIG_VALIDATION_SPEC,
        )
        assert result.ades_subindic == AdESFailure.REVOKED


@pytest.mark.asyncio
async def test_embedded_cades_revoked_no_poe(requests_mock):
    with freeze_time('2020-11-01'):
        signature = await REVOKED_SIGNER.async_sign_general_data(
            b'Hello world!',
            'sha256',
            detached=False,
            timestamper=DUMMY_TS,
            use_cades=True,
            signed_attr_settings=PdfCMSSignedAttributes(
                cades_signed_attrs=CAdESSignedAttrSpec(timestamp_content=False)
            ),
        )
        signature = cms.ContentInfo.load(signature.dump())

    with freeze_time('2020-12-25'):
        live_testing_vc(requests_mock)
        result = await ades.ades_basic_validation(
            signature['content'],
            validation_spec=DEFAULT_SIG_VALIDATION_SPEC,
        )
        assert result.ades_subindic == AdESIndeterminate.REVOKED_NO_POE


@pytest.mark.asyncio
async def test_embedded_cades_pre_revoke_with_poe(requests_mock):
    with freeze_time('2020-11-01'):
        signature = await REVOKED_SIGNER.async_sign_general_data(
            b'Hello world!',
            'sha256',
            detached=False,
            timestamper=DUMMY_TS,
            use_cades=True,
        )
        signature = cms.ContentInfo.load(signature.dump())
        vc = live_testing_vc(requests_mock)

        await async_validate_path(
            vc,
            ValidationPath(
                trust_anchor=CertTrustAnchor(TRUST_ROOTS[0]),
                interm=[INTERM_CERT],
                leaf=REVOKED_SIGNER.signing_cert,
            ),
        )

    with freeze_time('2020-12-25'):
        result = await ades.ades_with_time_validation(
            signature['content'],
            validation_spec=DEFAULT_SIG_VALIDATION_SPEC,
        )
        assert result.ades_subindic == AdESPassed.OK
        assert result.best_signature_time == datetime.datetime(
            2020, 11, 1, tzinfo=datetime.timezone.utc
        )


@pytest.mark.asyncio
@freeze_time('2020-11-20')
async def test_pades_lta_happy_path_current_time(requests_mock):
    out = await _generate_pades_test_doc(requests_mock)
    r = PdfFileReader(out)
    result = await ades.ades_lta_validation(
        r.embedded_signatures[0],
        pdf_validation_spec=DEFAULT_PDF_VALIDATION_SPEC,
    )
    assert result.ades_subindic == AdESPassed.OK


@pytest.mark.asyncio
async def test_pades_lta_happy_path_past_time(requests_mock):
    with freeze_time('2020-11-20'):
        out = await _generate_pades_test_doc(requests_mock)

    with freeze_time('2021-11-20'):
        r = PdfFileReader(out)
        result = await ades.ades_lta_validation(
            r.embedded_signatures[0],
            pdf_validation_spec=DEFAULT_PDF_VALIDATION_SPEC,
        )
        assert result.ades_subindic == AdESPassed.OK
        assert result.best_signature_time == datetime.datetime(
            2020, 11, 20, tzinfo=datetime.timezone.utc
        )


@pytest.mark.asyncio
@pytest.mark.parametrize('with_lta', [True, False])
@freeze_time('2020-11-20')
async def test_simulate_future_lta_happy_path(requests_mock, with_lta):
    # it shouldn't matter for the purposes of this function whether the initial
    # signature is followed by a DTS or not, so we test both
    out = await _generate_pades_test_doc(requests_mock, use_pades_lta=with_lta)

    r = PdfFileReader(out)
    result = await ades.simulate_future_ades_lta_validation(
        r.embedded_signatures[0],
        pdf_validation_spec=DEFAULT_PDF_VALIDATION_SPEC,
        future_validation_time=datetime.datetime(
            2030, 11, 20, tzinfo=datetime.timezone.utc
        ),
    )
    assert result.ades_subindic == AdESPassed.OK
    assert result.best_signature_time == datetime.datetime(
        2020, 11, 20, tzinfo=datetime.timezone.utc
    )


@pytest.mark.asyncio
async def test_simulate_future_lta_happy_path_with_ts_chain(requests_mock):
    with freeze_time('2020-11-20'):
        out = await _generate_pades_test_doc(requests_mock)

    with freeze_time('2028-11-20'):
        await _update_pades_test_doc(requests_mock, out)

        r = PdfFileReader(out)
        result = await ades.simulate_future_ades_lta_validation(
            r.embedded_signatures[0],
            pdf_validation_spec=DEFAULT_PDF_VALIDATION_SPEC,
            future_validation_time=datetime.datetime(
                2050, 11, 20, tzinfo=datetime.timezone.utc
            ),
        )
    assert result.ades_subindic == AdESPassed.OK
    assert result.best_signature_time == datetime.datetime(
        2020, 11, 20, tzinfo=datetime.timezone.utc
    )


@pytest.mark.asyncio
@freeze_time('2020-11-20')
async def test_simulate_future_lta_no_revinfo_fail(requests_mock):
    out = await _generate_pades_test_doc(
        requests_mock, embed_validation_info=False
    )
    r = PdfFileReader(out)
    result = await ades.simulate_future_ades_lta_validation(
        r.embedded_signatures[0],
        pdf_validation_spec=DEFAULT_PDF_VALIDATION_SPEC,
        future_validation_time=datetime.datetime(
            2030, 11, 20, tzinfo=datetime.timezone.utc
        ),
    )
    assert result.ades_subindic.status == AdESStatus.INDETERMINATE


@pytest.mark.asyncio
async def test_simulate_future_lta_with_broken_ts_chain(requests_mock):
    with freeze_time('2020-11-20'):
        out = await _generate_pades_test_doc(requests_mock)

    # gap too large
    with freeze_time('2031-11-20'):
        w = IncrementalPdfFileWriter(out)
        vc = live_testing_vc(requests_mock)
        await PdfTimeStamper(DUMMY_TS2).async_timestamp_pdf(
            w, md_algorithm='sha256', validation_context=vc, in_place=True
        )

        r = PdfFileReader(out)
        result = await ades.simulate_future_ades_lta_validation(
            r.embedded_signatures[0],
            pdf_validation_spec=DEFAULT_PDF_VALIDATION_SPEC,
            future_validation_time=datetime.datetime(
                2050, 11, 20, tzinfo=datetime.timezone.utc
            ),
        )
    assert result.ades_subindic == AdESIndeterminate.OUT_OF_BOUNDS_NO_POE


@pytest.mark.asyncio
@pytest.mark.parametrize('with_dts', [True, False])
async def test_pades_lta_expired_timestamp(requests_mock, with_dts):
    with freeze_time('2020-11-20'):
        out = await _generate_pades_test_doc(
            requests_mock, use_pades_lta=with_dts
        )

    with freeze_time('2080-11-20'):
        r = PdfFileReader(out)
        result = await ades.ades_lta_validation(
            r.embedded_signatures[0],
            pdf_validation_spec=DEFAULT_PDF_VALIDATION_SPEC,
        )
        assert result.ades_subindic == AdESIndeterminate.NO_POE


@pytest.mark.asyncio
async def test_pades_lta_happy_path_past_time_with_chain(requests_mock):
    with freeze_time('2020-11-20'):
        out = await _generate_pades_test_doc(requests_mock)

    with freeze_time('2028-11-20'):
        await _update_pades_test_doc(requests_mock, out)

    with freeze_time('2035-11-20'):
        r = PdfFileReader(out)
        result = await ades.ades_lta_validation(
            r.embedded_signatures[0],
            pdf_validation_spec=DEFAULT_PDF_VALIDATION_SPEC,
        )
        assert result.ades_subindic == AdESPassed.OK
        assert result.best_signature_time == datetime.datetime(
            2020, 11, 20, tzinfo=datetime.timezone.utc
        )


class NoSha512AfterSomeTime(AlgorithmUsagePolicy):
    # Algo policy for tests

    def __init__(self, year):
        self.cutoff = datetime.datetime(
            year, 12, 31, 23, 59, 59, tzinfo=datetime.timezone.utc
        )

    def signature_algorithm_allowed(
        self,
        algo: algos.SignedDigestAlgorithm,
        moment: Optional[datetime.datetime],
        public_key: Optional[keys.PublicKeyInfo],
    ) -> AlgorithmUsageConstraint:
        try:
            h = algo.hash_algo
        except ValueError:
            h = None

        if h == 'sha512':
            if moment is None or moment > self.cutoff:
                return AlgorithmUsageConstraint(
                    allowed=False,
                    not_allowed_after=self.cutoff,
                    failure_reason='just because',
                )
        return AlgorithmUsageConstraint(allowed=True)

    def digest_algorithm_allowed(
        self, algo: algos.DigestAlgorithm, moment: Optional[datetime.datetime]
    ) -> AlgorithmUsageConstraint:
        if algo['algorithm'].native == 'sha512':
            if moment is None or moment > self.cutoff:
                return AlgorithmUsageConstraint(
                    allowed=False,
                    not_allowed_after=self.cutoff,
                    failure_reason='just because',
                )
        return AlgorithmUsageConstraint(allowed=True)


class BanAllTheThings(AlgorithmUsagePolicy):
    def signature_algorithm_allowed(
        self,
        algo: algos.SignedDigestAlgorithm,
        moment: Optional[datetime.datetime],
        public_key: Optional[keys.PublicKeyInfo],
    ) -> AlgorithmUsageConstraint:
        return AlgorithmUsageConstraint(allowed=False)

    def digest_algorithm_allowed(
        self, algo: algos.DigestAlgorithm, moment: Optional[datetime.datetime]
    ) -> AlgorithmUsageConstraint:
        return AlgorithmUsageConstraint(allowed=False)


def _assert_certs_known(certs):
    return [
        KnownPOE(
            digest=digest_for_poe(cert.dump()),
            poe_time=cert.not_valid_before,
            poe_type=POEType.PROVIDED,
            validation_object=ValidationObject(
                object_type=ValidationObjectType.CERTIFICATE,
                value=cert,
            ),
        )
        for cert in certs
    ]


@pytest.mark.parametrize('place', ['in_sig', 'in_cert', 'both'])
@pytest.mark.asyncio
async def test_pades_hash_algorithm_banned_but_poe_ok(requests_mock, place):
    md_algorithm = 'sha256'
    signer = FROM_CA
    if place == 'in_sig' or place == 'both':
        md_algorithm = 'sha512'
    if place == 'in_cert' or place == 'both':
        store = SimpleCertificateStore().from_certs(FROM_CA.cert_registry)
        signer = SimpleSigner(
            signing_cert=TESTING_CA.get_cert('signer1-sha512'),
            signing_key=FROM_CA.signing_key,
            cert_registry=store,
        )
    with freeze_time('2020-11-20'):
        out = await _generate_pades_test_doc(
            requests_mock, md_algorithm=md_algorithm, signer=signer
        )

    with freeze_time('2029-11-20'):
        revinfo_policy = (
            DEFAULT_SIG_VALIDATION_SPEC.cert_validation_policy.revinfo_policy
        )
        spec = SignatureValidationSpec(
            cert_validation_policy=CertValidationPolicySpec(
                trust_manager=SimpleTrustManager.build(TRUST_ROOTS),
                algorithm_usage_policy=NoSha512AfterSomeTime(2025),
                revinfo_policy=revinfo_policy,
            ),
            local_knowledge=LocalKnowledge(
                known_poes=_assert_certs_known(FROM_CA.cert_registry)
            ),
        )
        r = PdfFileReader(out)
        result = await ades.ades_lta_validation(
            r.embedded_signatures[0],
            pdf_validation_spec=PdfSignatureValidationSpec(spec),
        )
        assert result.ades_subindic == AdESPassed.OK


@pytest.mark.parametrize(
    'place',
    ['in_sig', 'in_cert', 'both'],
)
@pytest.mark.asyncio
async def test_pades_lta_hash_algorithm_banned_and_no_poe(requests_mock, place):
    md_algorithm = 'sha256'
    signer = FROM_CA
    if place == 'in_sig' or place == 'both':
        md_algorithm = 'sha512'
    if place == 'in_cert' or place == 'both':
        # disable docts so we don't have PoE for the certs
        store = SimpleCertificateStore().from_certs(FROM_CA.cert_registry)
        signer = SimpleSigner(
            signing_cert=TESTING_CA.get_cert('signer1-sha512'),
            signing_key=FROM_CA.signing_key,
            cert_registry=store,
        )
    with freeze_time('2020-11-20'):
        out = await _generate_pades_test_doc(
            requests_mock, md_algorithm=md_algorithm, signer=signer
        )

    with freeze_time('2029-11-20'):
        revinfo_policy = (
            DEFAULT_SIG_VALIDATION_SPEC.cert_validation_policy.revinfo_policy
        )
        spec = SignatureValidationSpec(
            cert_validation_policy=CertValidationPolicySpec(
                trust_manager=SimpleTrustManager.build(TRUST_ROOTS),
                algorithm_usage_policy=NoSha512AfterSomeTime(2019),
                revinfo_policy=revinfo_policy,
            ),
            local_knowledge=LocalKnowledge(
                known_poes=_assert_certs_known(FROM_CA.cert_registry)
            ),
        )
        r = PdfFileReader(out)
        result = await ades.ades_lta_validation(
            r.embedded_signatures[0],
            pdf_validation_spec=PdfSignatureValidationSpec(spec),
        )
        assert (
            result.ades_subindic
            == AdESIndeterminate.CRYPTO_CONSTRAINTS_FAILURE_NO_POE
        )


@pytest.mark.asyncio
async def test_pades_lta_algo_permaban(requests_mock):
    with freeze_time('2020-11-20'):
        out = await _generate_pades_test_doc(
            requests_mock, md_algorithm='sha512'
        )

    with freeze_time('2029-11-20'):
        revinfo_policy = (
            DEFAULT_SIG_VALIDATION_SPEC.cert_validation_policy.revinfo_policy
        )
        spec = SignatureValidationSpec(
            cert_validation_policy=CertValidationPolicySpec(
                trust_manager=SimpleTrustManager.build(TRUST_ROOTS),
                algorithm_usage_policy=BanAllTheThings(),
                revinfo_policy=revinfo_policy,
            )
        )
        r = PdfFileReader(out)
        result = await ades.ades_lta_validation(
            r.embedded_signatures[0],
            pdf_validation_spec=PdfSignatureValidationSpec(spec),
        )
        assert (
            result.ades_subindic == AdESIndeterminate.CRYPTO_CONSTRAINTS_FAILURE
        )


@pytest.mark.asyncio
async def test_pades_lta_live_ac_validation(requests_mock):
    with freeze_time('2020-11-01'):
        pki_arch = CERTOMANCER.get_pki_arch(ArchLabel('testing-ca-with-aa'))
        authorities = [
            pki_arch.get_cert('root'),
            pki_arch.get_cert('interm'),
            pki_arch.get_cert('root-aa'),
            pki_arch.get_cert('interm-aa'),
            pki_arch.get_cert('leaf-aa'),
        ]
        signer = signers.SimpleSigner(
            signing_cert=pki_arch.get_cert(CertLabel('signer1')),
            signing_key=pki_arch.key_set.get_private_key(KeyLabel('signer1')),
            cert_registry=SimpleCertificateStore.from_certs(authorities),
            attribute_certs=[
                pki_arch.get_attr_cert(CertLabel('alice-role-with-rev'))
            ],
        )
        dummy_ts = timestamps.DummyTimeStamper(
            tsa_cert=pki_arch.get_cert(CertLabel('tsa')),
            tsa_key=pki_arch.key_set.get_private_key(KeyLabel('tsa')),
            certs_to_embed=SimpleCertificateStore.from_certs(
                [pki_arch.get_cert('root')]
            ),
        )

        fetchers = RequestsFetcherBackend().get_fetchers()
        vc = ValidationContext(
            trust_roots=[pki_arch.get_cert('root')],
            allow_fetching=True,
            other_certs=authorities,
            fetchers=fetchers,
            revocation_mode='require',
        )
        ac_vc = ValidationContext(
            trust_roots=[pki_arch.get_cert('root-aa')],
            allow_fetching=True,
            other_certs=authorities,
            fetchers=fetchers,
            revocation_mode='require',
        )

        Illusionist(pki_arch).register(requests_mock)

        w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
        out = await signers.async_sign_pdf(
            w,
            signers.PdfSignatureMetadata(
                validation_context=vc,
                ac_validation_context=ac_vc,
                subfilter=PADES,
                embed_validation_info=True,
                use_pades_lta=True,
            ),
            signer=signer,
            timestamper=dummy_ts,
            existing_fields_only=True,
        )

    revinfo_policy = certv_policy_decl.CertRevTrustPolicy(
        revocation_checking_policy=certv_policy_decl.REQUIRE_REVINFO,
        freshness_req_type=FreshnessReqType.MAX_DIFF_REVOCATION_VALIDATION,
    )
    sig_validation_spec = SignatureValidationSpec(
        cert_validation_policy=CertValidationPolicySpec(
            trust_manager=SimpleTrustManager.build([pki_arch.get_cert('root')]),
            revinfo_policy=revinfo_policy,
        ),
        ac_validation_policy=CertValidationPolicySpec(
            trust_manager=SimpleTrustManager.build(
                [pki_arch.get_cert('root-aa')]
            ),
            revinfo_policy=revinfo_policy,
        ),
    )
    with freeze_time('2028-02-01'):
        r = PdfFileReader(out)
        result = await ades.ades_lta_validation(
            r.embedded_signatures[0],
            pdf_validation_spec=PdfSignatureValidationSpec(sig_validation_spec),
        )
        assert result.ades_subindic == AdESPassed.OK

        roles = list(result.api_status.ac_attrs['role'].attr_values)
        role = roles[0]
        assert isinstance(role, cms.RoleSyntax)
        assert len(result.api_status.ac_attrs) == 1
        assert role['role_name'].native == 'bigboss@example.com'


async def _nontraditional_hybrid_lta_doc(requests_mock):
    # this document reproduces the situation of #228:
    #  - No DSS or DTSes
    #  - declared PAdES (/ETSI.CAdES.detached)
    #  - short-lived leaf cert (=> expired by validation time)
    #  - Leaf cert OCSP response in Adobe revinfo archival
    #  - Revinfo for intermediate cert must be fetched
    #  - Different TS root.
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    vc = live_testing_vc(requests_mock, with_extra_tsa=True)

    custom_signer = signers.SimpleSigner(
        signing_cert=FROM_CA.signing_cert,
        signing_key=FROM_CA.signing_key,
        cert_registry=FROM_CA.cert_registry,
    )
    cert_id = ocsp.CertId(
        {
            'hash_algorithm': algos.DigestAlgorithm({'algorithm': 'sha256'}),
            'issuer_name_hash': TESTING_CA.entities.get_name_hash(
                EntityLabel('interm'), 'sha256'
            ),
            'issuer_key_hash': INTERM_CERT.public_key.sha256,
            'serial_number': FROM_CA.signing_cert.serial_number,
        }
    )
    responder = TESTING_CA.service_registry.summon_responder(
        ServiceLabel('interm'),
        at_time=datetime.datetime.now(tz=tzlocal.get_localzone()),
    )
    ocsp_response = responder.assemble_simple_ocsp_responses(
        [responder.format_single_ocsp_response(cert_id, INTERM_CERT)]
    )
    custom_signer.signed_attr_prov_spec = GenericPdfSignedAttributeProviderSpec(
        attr_settings=PdfCMSSignedAttributes(
            adobe_revinfo_attr=RevocationInfoArchival({'ocsp': [ocsp_response]})
        ),
        signing_cert=custom_signer.signing_cert,
        signature_mechanism=custom_signer.signature_mechanism,
        timestamper=None,
    )

    timestamper = DUMMY_HTTP_TS_VARIANT
    out = await signers.async_sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            validation_context=vc,
            subfilter=PADES,
            embed_validation_info=False,  # done manually
        ),
        signer=custom_signer,
        timestamper=timestamper,
    )
    r = PdfFileReader(out)
    assert len(r.embedded_signatures) == 1
    assert '/DSS' not in r.root
    return out


@pytest.mark.asyncio
async def test_nontraditional_hybrid_lta(requests_mock):
    with freeze_time('2020-11-20'):
        out = await _nontraditional_hybrid_lta_doc(requests_mock)

    modified_policy = SignatureValidationSpec(
        cert_validation_policy=CertValidationPolicySpec(
            trust_manager=SimpleTrustManager.build(
                TRUST_ROOTS + [UNRELATED_TSA.get_cert('root')]
            ),
            revinfo_policy=DEFAULT_REVINFO_POLICY,
        )
    )
    with freeze_time('2022-11-20'):
        r = PdfFileReader(out)
        result = await ades.ades_lta_validation(
            r.embedded_signatures[0],
            pdf_validation_spec=PdfSignatureValidationSpec(modified_policy),
        )
        assert result.ades_subindic == AdESPassed.OK
        assert result.api_status.bottom_line
        assert result.api_status.coverage == SignatureCoverageLevel.ENTIRE_FILE
        assert result.best_signature_time == datetime.datetime(
            2020, 11, 20, tzinfo=datetime.timezone.utc
        )


@pytest.mark.asyncio
async def test_nontraditional_hybrid_lta_with_failed_timestamp(requests_mock):
    with freeze_time('2020-11-20'):
        out = await _nontraditional_hybrid_lta_doc(requests_mock)

    with freeze_time('2022-11-20'):
        r = PdfFileReader(out)
        result = await ades.ades_lta_validation(
            r.embedded_signatures[0],
            pdf_validation_spec=DEFAULT_PDF_VALIDATION_SPEC,
        )
        # since we assert that time stamp tokens must be valid, the
        # spec requires us to return the TS validation result
        assert (
            result.ades_subindic == AdESIndeterminate.NO_CERTIFICATE_CHAIN_FOUND
        )
        assert not result.api_status.bottom_line
        assert result.api_status.coverage == SignatureCoverageLevel.ENTIRE_FILE
        assert result.best_signature_time == datetime.datetime(
            2022, 11, 20, tzinfo=datetime.timezone.utc
        )



from .commons import qualify
from .form_rules_api import (
    FieldComparisonContext,
    FieldComparisonSpec,
    FieldMDPRule,
    FormUpdate,
    FormUpdatingRule,
)
from .policies import (
    DEFAULT_DIFF_POLICY,
    NO_CHANGES_DIFF_POLICY,
    StandardDiffPolicy,
)
from .policy_api import (
    DiffPolicy,
    DiffResult,
    ModificationLevel,
    SuspiciousModification,
)
from .rules.file_structure_rules import (
    CatalogModificationRule,
    ObjectStreamRule,
    XrefStreamRule,
)
from .rules.form_field_rules import (
    BaseFieldModificationRule,
    DSSCompareRule,
    GenericFieldModificationRule,
    SigFieldCreationRule,
    SigFieldModificationRule,
)
from .rules.metadata_rules import DocInfoRule, MetadataUpdateRule
from .rules_api import QualifiedWhitelistRule, ReferenceUpdate, WhitelistRule

# export list for compatibility with the old module

__all__ = [
    'ModificationLevel',
    'SuspiciousModification',
    'QualifiedWhitelistRule',
    'WhitelistRule',
    'qualify',
    'ReferenceUpdate',
    'DocInfoRule',
    'DSSCompareRule',
    'MetadataUpdateRule',
    'CatalogModificationRule',
    'ObjectStreamRule',
    'XrefStreamRule',
    'FormUpdatingRule',
    'FormUpdate',
    'FieldMDPRule',
    'FieldComparisonSpec',
    'FieldComparisonContext',
    'GenericFieldModificationRule',
    'SigFieldCreationRule',
    'SigFieldModificationRule',
    'BaseFieldModificationRule',
    'DiffPolicy',
    'StandardDiffPolicy',
    'DEFAULT_DIFF_POLICY',
    'NO_CHANGES_DIFF_POLICY',
    'DiffResult',
]



import binascii
import uuid
from enum import Enum
from typing import Optional

from .generic import (
    DictionaryObject,
    NameObject,
    PdfObject,
    StreamObject,
    pdf_name,
)
from .layout import BoxConstraints
from .reader import PdfFileReader
from .writer import BasePdfFileWriter

__all__ = [
    'ResourceType',
    'ResourceManagementError',
    'PdfResources',
    'PdfContent',
    'RawContent',
    'ImportedPdfPage',
]

# TODO have the merge_resources helper in incremental_writer rely on some
#  of the idioms established here


class ResourceType(Enum):
    """
    Enum listing resources that can be used as keys in a resource dictionary.

    See ISO 32000-1, § 7.8.3 Table 34.
    """

    EXT_G_STATE = pdf_name('/ExtGState')
    """
    External graphics state specifications.
    See ISO 32000-1, § 8.4.5.
    """

    COLOR_SPACE = pdf_name('/ColorSpace')
    """
    Colour space definitions.
    See ISO 32000-1, § 8.6.
    """

    PATTERN = pdf_name('/Pattern')
    """
    Pattern definitions.
    See ISO 32000-1, § 8.7.
    """

    SHADING = pdf_name('/Shading')
    """
    Shading definitions.
    See ISO 32000-1, § 8.7.4.3.
    """

    XOBJECT = pdf_name('/XObject')
    """
    External object definitions (images and form XObjects).
    See ISO 32000-1, § 8.8.
    """

    FONT = pdf_name('/Font')
    """
    Font specifications.
    See ISO 32000-1, § 9.
    """

    PROPERTIES = pdf_name('/Properties')
    """
    Marked content properties.
    See ISO 32000-1, § 14.6.2.
    """


class ResourceManagementError(ValueError):
    """
    Used to signal problems with resource dictionaries.
    """

    pass


def _res_merge_helper(dict1, dict2):
    for k, v2 in dict2.items():
        if k in dict1:
            raise ResourceManagementError(
                f"Resource with name {k} occurs in both dictionaries."
            )
        dict1[k] = v2
    return dict1


class PdfResources:
    """
    Representation of a PDF resource dictionary.

    This class implements :meth:`__getitem__` with :class:`.ResourceType` keys
    for dynamic access to its attributes.
    To merge two instances of :class:`.PdfResources` into one another,
    the class overrides :meth:`__iadd__`, so you can write.

    .. code-block:: python

        res1 += res2

    *Note:* Merging two resource dictionaries with conflicting resource names
    will produce a :class:`.ResourceManagementError`.

    *Note:* This class is currently only used for new resource dictionaries.
    """

    def __init__(self):
        self.ext_g_state = DictionaryObject()
        self.color_space = DictionaryObject()
        self.pattern = DictionaryObject()
        self.shading = DictionaryObject()
        self.xobject = DictionaryObject()
        self.font = DictionaryObject()
        self.properties = DictionaryObject()

    def __getitem__(self, item: ResourceType):
        return getattr(self, item.name.lower())

    def as_pdf_object(self) -> DictionaryObject:
        """
        Render this instance of :class:`.PdfResources` to an actual resource
        dictionary.
        """

        def _gen():
            for k in ResourceType:
                val = self[k]
                if val:
                    yield k.value, val

        return DictionaryObject({k: v for k, v in _gen()})

    def __iadd__(self, other):
        """
        Merge another resource dictionary into this one.
        :param other:
            Another instance of :class:`.PdfResources`
        :return:
            Always returns ``self``
        :raises ResourceManagementError:
            Raised when there is a resource name conflict.
        """
        for k in ResourceType:
            _res_merge_helper(self[k], other[k])
        return self


class PdfContent:
    """
    Abstract representation of part of a PDF content stream.

    .. warning::

        Whether :class:`.PdfContent` instances can be reused or not
        is left up to the subclasses.
    """

    writer = None
    """
    The :meth:`__init__` method comes with an optional ``writer`` 
    parameter that can be used to let subclasses register external resources 
    with the writer by themselves.
    
    It can also be set after the fact by calling :meth:`set_writer`.
    """

    def __init__(
        self,
        resources: Optional[PdfResources] = None,
        box: Optional[BoxConstraints] = None,
        writer: Optional[BasePdfFileWriter] = None,
    ):
        self._resources: PdfResources = resources or PdfResources()
        self.box: BoxConstraints = box or BoxConstraints()
        self.writer = writer

    @property
    def _ensure_writer(self) -> BasePdfFileWriter:
        if self.writer is None:
            raise ValueError("PDF writer is not set")
        return self.writer

    # TODO support a set-if-not-taken mechanism, that suggests alternative names
    #  if necessary.
    def set_resource(
        self, category: ResourceType, name: NameObject, value: PdfObject
    ):
        """Set a value in the resource dictionary associated with this content
        fragment.

        :param category:
            The resource category to which the resource belongs.
        :param name:
            The resource's (internal) name.
        :param value:
            The resource's value.
        """
        self._resources[category][name] = value

    def import_resources(self, resources: PdfResources):
        """Import resources from another resource dictionary.

        :param resources:
            An instance of :class:`.PdfResources`.
        :raises ResourceManagementError:
            Raised when there is a resource name conflict.
        """
        self._resources += resources

    @property
    def resources(self) -> PdfResources:
        """
        :return:
            The :class:`.PdfResources` instance associated with this
            content fragment.
        """
        return self._resources

    def render(self) -> bytes:
        """
        Compile the content to graphics operators.
        """
        raise NotImplementedError

    # TODO allow the bounding box to be overridden/refitted
    #  (using matrix transforms)
    def as_form_xobject(self) -> StreamObject:
        """
        Render the object to a form XObject to be referenced by another
        content stream. See ISO 32000-1, § 8.8.

        *Note:* Even if :attr:`writer` is set, the resulting form XObject will
        not be registered. This is left up to the caller.

        :return:
            A :class:`~.generic.StreamObject` instance representing
            the resulting form XObject.
        """
        from pyhanko.pdf_utils.writer import init_xobject_dictionary

        command_stream = self.render()
        return init_xobject_dictionary(
            command_stream=command_stream,
            box_width=self.box.width,
            box_height=self.box.height,
            resources=self._resources.as_pdf_object(),
        )

    def set_writer(self, writer):
        """
        Override the currently registered writer object.

        :param writer:
            An instance of :class:`~.writer.BasePdfFileWriter`.
        """
        self.writer = writer

    def add_to_page(
        self, writer: BasePdfFileWriter, page_ix: int, prepend: bool = False
    ):
        """
        Convenience wrapper around :meth:`.BasePdfFileWriter.add_stream_to_page`
        to turn a :class:`.PdfContent` instance into a page
        content stream.

        :param writer:
            A PDF file writer.
        :param page_ix:
            Index of the page to modify.
            The first page has index `0`.
        :param prepend:
            Prepend the content stream to the list of content streams, as
            opposed to appending it to the end.
            This has the effect of causing the stream to be rendered
            underneath the already existing content on the page.
        :return:
            An :class:`~.generic.IndirectObject` reference to the page object
            that was modified.
        """
        as_stream = StreamObject({}, stream_data=self.render())
        return writer.add_stream_to_page(
            page_ix,
            writer.add_object(as_stream),
            resources=self.resources.as_pdf_object(),
            prepend=prepend,
        )


class RawContent(PdfContent):
    """Raw byte sequence to be used as PDF content."""

    def __init__(
        self,
        data: bytes,
        resources: Optional[PdfResources] = None,
        box: Optional[BoxConstraints] = None,
    ):
        super().__init__(resources, box)
        self.data = data

    def render(self) -> bytes:
        return self.data


class ImportedPdfPage(PdfContent):
    """Import a page from another PDF file (lazily)"""

    def __init__(self, file_name, page_ix=0):
        self.file_name = file_name
        self.page_ix = page_ix
        super().__init__()

    def render(self) -> bytes:
        from .writer import BasePdfFileWriter

        w: BasePdfFileWriter = self._ensure_writer
        with open(self.file_name, 'rb') as inf:
            r = PdfFileReader(inf)
            xobj = w.import_page_as_xobject(r, page_ix=self.page_ix)
        resource_name = b'/Import' + binascii.hexlify(uuid.uuid4().bytes)
        self.resources.xobject[resource_name.decode('ascii')] = xobj

        # make sure to take the bounding box (i.e. the page's MediaBox)
        # into account when doing layout computations
        x1, y1, x2, y2 = xobj.get_object()['/BBox']
        self.box = BoxConstraints(width=abs(x1 - x2), height=abs(y1 - y2))
        return resource_name + b' Do'



"""
Tests for PDF MAC, AKA ISO 32004.

Note: most of the failure cases follow "known bad" sample files.
"""

import asyncio
import enum
import functools
import hashlib
import itertools
import os
import re
import typing
from functools import wraps
from io import BytesIO
from typing import Callable, Optional, Type

import pytest
from asn1crypto import core
from certomancer.integrations.illusionist import Illusionist
from freezegun import freeze_time
from pyhanko.pdf_utils import generic, writer
from pyhanko.pdf_utils.crypt import (
    PubKeySecurityHandler,
    StandardSecurityHandler,
    pdfmac,
)
from pyhanko.pdf_utils.crypt.api import PdfMacStatus
from pyhanko.pdf_utils.font.basic import get_courier
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
from pyhanko.pdf_utils.reader import PdfFileReader
from pyhanko.pdf_utils.writer import PdfFileWriter
from pyhanko.sign import (
    PdfSigner,
    PdfTimeStamper,
    SimpleSigner,
    signers,
    timestamps,
)
from pyhanko.sign.general import SigningError, simple_cms_attribute
from pyhanko.sign.signers import cms_embedder
from pyhanko.sign.signers.pdf_byterange import SigByteRangeObject
from pyhanko.sign.validation import (
    RevocationInfoValidationType,
    validate_pdf_ltv_signature,
)
from pyhanko.sign.validation.errors import DisallowedAlgorithmError
from requests_mock import Mocker

from pyhanko_certvalidator import ValidationContext

from .samples import *
from .signing_commons import (
    DUMMY_TS,
    ECC_ROOT_CERT,
    FROM_ECC_CA,
    SIMPLE_ECC_V_CONTEXT,
    val_trusted,
)

DUMMY_PASSWORD = "secret"


GENERATED_TEST_OUTPUTS: Optional[str] = None


def _dummy_decrypt(r: PdfFileReader, ignore_mac=False):
    if ignore_mac:
        r._validate_pdf_mac = id
    if isinstance(r.security_handler, PubKeySecurityHandler):
        result = r.decrypt_pubkey(PUBKEY_TEST_DECRYPTER)
    elif isinstance(r.security_handler, StandardSecurityHandler):
        result = r.decrypt(DUMMY_PASSWORD)
    else:
        raise NotImplementedError
    return result


def _fmt_arg(arg: typing.Any) -> str:
    # hacky helper function for choosing file names for test outputs
    if isinstance(arg, (int, bool)):
        return str(arg)
    elif isinstance(arg, str) and arg.isalnum():
        return arg
    elif isinstance(arg, enum.Enum):
        return f"{arg.__class__.__name__}_{arg.name}"
    elif isinstance(arg, Mocker):
        return ""  # ignore
    else:
        raise NotImplementedError(
            f"Test with bad argument of type {type(arg)}; "
            f"can't derive output file name"
        )


def _preserve_test_result(
    subfolder: str, name: str, output: BytesIO, test_kwargs: dict
):
    """
    Hook to preserve test outputs as reference files.
    """

    if not GENERATED_TEST_OUTPUTS:
        return

    out_dir = os.path.join(GENERATED_TEST_OUTPUTS, subfolder)
    fmt_args = (_fmt_arg(v) for v in test_kwargs.values())
    name_suffix = '-'.join(p for p in fmt_args if p)
    if name_suffix:
        out_dir = os.path.join(out_dir, name)
        fname = f"{name}-{name_suffix}"
    else:
        fname = name
    os.makedirs(out_dir, exist_ok=True)
    with open(os.path.join(out_dir, f"{fname}.pdf"), 'wb') as outf:
        outf.write(output.getvalue())


def pdf_mac_good(test_fun):
    @wraps(test_fun)
    def _wrap(*args, **kwargs):
        output_buf = test_fun(*args, **kwargs)
        r = PdfFileReader(output_buf)
        result = _dummy_decrypt(r)
        assert result.mac_status == PdfMacStatus.SUCCESSFUL
        iso_exts = {
            int(ext.get_object()['/ExtensionLevel'])
            for ext in r.root['/Extensions']['/ISO_']
        }
        assert 32004 in iso_exts
        _preserve_test_result("good", test_fun.__name__, output_buf, kwargs)

    return _wrap


def pdfmac_known_bad_case(error_message: str):
    def dec(test_fun):
        @wraps(test_fun)
        def _wrap(*args, **kwargs):
            output_buf = test_fun(*args, **kwargs)
            r = PdfFileReader(output_buf)
            result = _dummy_decrypt(r)
            assert result.mac_status == PdfMacStatus.FAILED
            match = re.search(
                error_message, result.mac_failure_reason, re.IGNORECASE
            )
            assert match is not None, (
                f"Error {result.mac_failure_reason!r} did "
                f"not match {error_message!r}"
            )
            _preserve_test_result("bad", test_fun.__name__, output_buf, kwargs)

        return _wrap

    return dec


def shallow_wraps(wrapped):
    # same as functools.wraps() but doesn't expose the arguments list
    # of the underlying function by omitting __wrapped__
    # (and the annotation list).
    # Use case: create canned parameterised tests using decorators
    # (since pytest.mark.parameterized is a little too clever with arglist
    # introspection)

    def dec(wrapper):
        assigned = ('__module__', '__name__', '__qualname__', '__doc__')
        updated = functools.WRAPPER_UPDATES
        for attr in assigned:
            try:
                value = getattr(wrapped, attr)
            except AttributeError:
                pass
            else:
                setattr(wrapper, attr, value)
        for attr in updated:
            getattr(wrapper, attr).update(getattr(wrapped, attr, {}))
        return wrapper

    return dec


# This is an enum to improve the readability of pytest output


class MacLocation(enum.Enum):
    STANDALONE = 0
    IN_SIG = 1


class EncryptionType(enum.Enum):
    STANDARD = 0
    PUBKEY = 1


def init_sample_doc(
    encryption_type: EncryptionType,
    writer_class: Type[PdfFileWriter] = PdfFileWriter,
):
    w = writer_class(stream_xrefs=False)

    resources = generic.DictionaryObject(
        {
            generic.pdf_name('/Font'): generic.DictionaryObject(
                {generic.pdf_name('/F1'): get_courier(w)}
            )
        }
    )

    stream_content = '''
    BT
        /F1 12 Tf 40 700 Td 12 TL
        (Hello ISO/TS 32004!) Tj
    ET
    '''

    stream = generic.StreamObject(stream_data=stream_content.encode('latin1'))
    explanation_page = writer.PageObject(
        contents=w.add_object(stream),
        media_box=(0, 0, 595.28, 841.89),
        resources=resources,
    )
    w.insert_page(explanation_page)

    if encryption_type == EncryptionType.PUBKEY:
        sh = PubKeySecurityHandler.build_from_certs(
            [PUBKEY_TEST_DECRYPTER.cert], compat_entries=False
        )
        w._assign_security_handler(sh)
    else:
        w.encrypt(DUMMY_PASSWORD, compat_entries=False)
    return w


def pdfmac_with_handler(
    error_message: str, only_location: Optional[MacLocation] = None
):
    """
    Decorator to set up a test (or multiple tests) based on an alternative
    Handler implementation. Used to exercise the validator.
    """

    if only_location is not None:
        locations = [only_location]
    else:
        locations = iter(MacLocation)

    def dec(test_fun):
        @pytest.mark.parametrize(
            'location,encryption_type',
            itertools.product(locations, iter(EncryptionType)),
        )
        @pdfmac_known_bad_case(error_message)
        @shallow_wraps(test_fun)
        def _wrap(location: MacLocation, encryption_type: EncryptionType):
            handler_cls: Type[pdfmac.PdfMacTokenHandler] = test_fun()
            w = init_sample_doc(encryption_type)
            w._mac_handler_cls = handler_cls

            if location == MacLocation.IN_SIG:
                signer = SimpleSigner(
                    signing_cert=FROM_ECC_CA.signing_cert,
                    signing_key=FROM_ECC_CA.signing_key,
                    cert_registry=FROM_ECC_CA.cert_registry,
                )
                out = signers.sign_pdf(
                    w,
                    signers.PdfSignatureMetadata(field_name='Sig'),
                    signer=signer,
                )
            else:
                out = BytesIO()
                w.write(out)

            return out

        return _wrap

    return dec


@pdf_mac_good
@pytest.mark.parametrize('encryption_type', list(EncryptionType))
def test_simple_mac(encryption_type):
    w = init_sample_doc(encryption_type)
    out = BytesIO()
    w.write(out)
    return out


@pdf_mac_good
@pytest.mark.parametrize('encryption_type', list(EncryptionType))
def test_simple_mac_nondefault_hash(encryption_type):
    w = init_sample_doc(encryption_type)

    class Handler(pdfmac.PdfMacTokenHandler):
        @classmethod
        def from_key_mat(
            cls, file_encryption_key: bytes, kdf_salt: bytes, md_algorithm: str
        ):
            mac_kek = cls._derive_mac_kek(file_encryption_key, kdf_salt)
            return cls(mac_kek=mac_kek, md_algorithm='sha384')

    w._mac_handler_cls = Handler
    out = BytesIO()
    w.write(out)
    return out


@freeze_time('2020-11-01')
@pdf_mac_good
@pytest.mark.parametrize('encryption_type', list(EncryptionType))
def test_sign_crypt_aes256_with_mac(encryption_type):
    w = init_sample_doc(encryption_type)
    signer = SimpleSigner(
        signing_cert=FROM_ECC_CA.signing_cert,
        signing_key=FROM_ECC_CA.signing_key,
        cert_registry=FROM_ECC_CA.cert_registry,
    )
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(field_name='Sig'),
        signer=signer,
    )

    r = PdfFileReader(out)
    _dummy_decrypt(r)
    s = r.embedded_signatures[0]
    val_trusted(s, vc=SIMPLE_ECC_V_CONTEXT())
    return out


@pdf_mac_good
@pytest.mark.parametrize('encryption_type', list(EncryptionType))
def test_signature_nondefault_hash(encryption_type):
    w = init_sample_doc(encryption_type)
    signer = SimpleSigner(
        signing_cert=FROM_ECC_CA.signing_cert,
        signing_key=FROM_ECC_CA.signing_key,
        cert_registry=FROM_ECC_CA.cert_registry,
    )
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(field_name='Sig', md_algorithm='sha384'),
        signer=signer,
    )

    return out


@pdf_mac_good
@freeze_time('2020-11-01')
def test_pdf_mac_pades(requests_mock):
    w = init_sample_doc(EncryptionType.STANDARD)

    trust_roots = [ECC_ROOT_CERT]
    vc = ValidationContext(
        trust_roots=trust_roots, allow_fetching=True, other_certs=[]
    )
    Illusionist(TESTING_CA_ECDSA).register(requests_mock)

    from .test_pades import PADES

    signer = SimpleSigner(
        signing_cert=FROM_ECC_CA.signing_cert,
        signing_key=FROM_ECC_CA.signing_key,
        cert_registry=FROM_ECC_CA.cert_registry,
    )
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            subfilter=PADES,
            validation_context=vc,
            embed_validation_info=True,
            use_pades_lta=True,
        ),
        timestamper=timestamps.HTTPTimeStamper(
            'http://pyhanko.tests/testing-ca-ecdsa/tsa/tsa',
        ),
        signer=signer,
    )

    r = PdfFileReader(out)
    r.decrypt(DUMMY_PASSWORD)
    validate_pdf_ltv_signature(
        r.embedded_signatures[0],
        validation_type=RevocationInfoValidationType.PADES_LTA,
        validation_context_kwargs={
            'trust_roots': trust_roots,
            'allow_fetching': False,
            'revocation_mode': 'soft-fail',
        },
    )
    return out


@pdf_mac_good
@pytest.mark.parametrize('encryption_type', list(EncryptionType))
@freeze_time('2020-11-01')
def test_pdf_mac_timestamp(encryption_type):
    w = init_sample_doc(EncryptionType.STANDARD)
    out = BytesIO()
    w.write(out)

    w = IncrementalPdfFileWriter(out)
    w.encrypt("secret")
    PdfTimeStamper(timestamper=DUMMY_TS).timestamp_pdf(
        w, md_algorithm='sha256', in_place=True
    )
    return out


@pdf_mac_good
@pytest.mark.parametrize('encryption_type', list(EncryptionType))
def test_cms_algorithm_protection_attr_missing(encryption_type):
    # this is actually allowed, it's a 'should' in ISO 32004

    class Handler(pdfmac.PdfMacTokenHandler):
        def _format_auth_attrs(
            self, message_digest: bytes
        ) -> cms.CMSAttributes:
            return cms.CMSAttributes(
                [
                    simple_cms_attribute(
                        'content_type', 'pdf_mac_integrity_info'
                    ),
                    simple_cms_attribute('message_digest', message_digest),
                ]
            )

    w = init_sample_doc(encryption_type)
    w._mac_handler_cls = Handler
    out = BytesIO()
    w.write(out)

    return out


@pdfmac_known_bad_case("Document digest does not match")
@pytest.mark.parametrize('encryption_type', list(EncryptionType))
def test_standalone_mac_tamper(encryption_type):
    w = init_sample_doc(encryption_type)
    out = BytesIO()
    w.write(out)
    # this is in the middle of the second line of the header, so pretty benign
    out.seek(10)
    out.write(b'\x01')
    return out


@pdfmac_with_handler("Failed to unwrap")
def test_alter_key():
    class Handler(pdfmac.PdfMacTokenHandler):
        def _get_mac_keying_info(self, dry_run):
            mac_key, ri = super()._get_mac_keying_info(dry_run)
            # by altering the key record, we mess up the unwrapping later
            ek = ri.chosen['encrypted_key']
            ri.chosen['encrypted_key'] = bytes(reversed(ek.native))
            return mac_key, ri

    return Handler


@pdfmac_with_handler("Failed to unwrap")
def test_alter_salt():
    class Handler(pdfmac.PdfMacTokenHandler):
        @classmethod
        def from_key_mat(
            cls, file_encryption_key: bytes, kdf_salt: bytes, md_algorithm: str
        ):
            return super().from_key_mat(
                file_encryption_key=file_encryption_key,
                # do calculations with a different salt
                kdf_salt=bytes(32),
                md_algorithm=md_algorithm,
            )

    return Handler


@pdfmac_with_handler("token has invalid MAC")
def test_append_auth_attr():
    class Handler(pdfmac.PdfMacTokenHandler):
        def compute_mac(self, mac_key: bytes, data_to_mac: bytes) -> bytes:
            # the point of this one is that the MAC was computed
            # over a different set of authenticated attributes,
            # so we might as well mess with the MAC computation instead
            return super().compute_mac(
                mac_key=mac_key, data_to_mac=cms.CMSAttributes([]).dump()
            )

    return Handler


@pdfmac_with_handler("Multiple CMS algorithm protection attributes present")
def test_cms_algorithm_protection_attr_multival():
    class Handler(pdfmac.PdfMacTokenHandler):
        def _format_auth_attrs(
            self, message_digest: bytes
        ) -> cms.CMSAttributes:
            algo_protection = cms.CMSAlgorithmProtection(
                {
                    'mac_algorithm': self.mac_algo_ident,
                    'digest_algorithm': cms.DigestAlgorithm(
                        {'algorithm': self.md_algorithm}
                    ),
                }
            )
            return cms.CMSAttributes(
                [
                    simple_cms_attribute(
                        'content_type', 'pdf_mac_integrity_info'
                    ),
                    simple_cms_attribute('message_digest', message_digest),
                    cms.CMSAttribute(
                        {
                            'type': cms.CMSAttributeType(
                                'cms_algorithm_protection'
                            ),
                            'values': (algo_protection, algo_protection),
                        }
                    ),
                ]
            )

    return Handler


@pdfmac_with_handler("duplicated")
def test_cms_algorithm_protection_attr_duplicated():
    class Handler(pdfmac.PdfMacTokenHandler):
        def _format_auth_attrs(
            self, message_digest: bytes
        ) -> cms.CMSAttributes:
            algo_protection = cms.CMSAlgorithmProtection(
                {
                    'mac_algorithm': self.mac_algo_ident,
                    'digest_algorithm': cms.DigestAlgorithm(
                        {'algorithm': self.md_algorithm}
                    ),
                }
            )
            return cms.CMSAttributes(
                [
                    simple_cms_attribute(
                        'content_type', 'pdf_mac_integrity_info'
                    ),
                    simple_cms_attribute('message_digest', message_digest),
                    simple_cms_attribute(
                        'cms_algorithm_protection', algo_protection
                    ),
                    simple_cms_attribute(
                        'cms_algorithm_protection', algo_protection
                    ),
                ]
            )

    return Handler


@pdfmac_with_handler("Digest algorithm does not match")
def test_cms_algorithm_protection_digest_algo_mismatch():
    class Handler(pdfmac.PdfMacTokenHandler):
        def _format_auth_attrs(
            self, message_digest: bytes
        ) -> cms.CMSAttributes:
            algo_protection = cms.CMSAlgorithmProtection(
                {
                    'mac_algorithm': self.mac_algo_ident,
                    'digest_algorithm': cms.DigestAlgorithm(
                        {'algorithm': 'sha3_512'}
                    ),
                }
            )
            return cms.CMSAttributes(
                [
                    simple_cms_attribute(
                        'content_type', 'pdf_mac_integrity_info'
                    ),
                    simple_cms_attribute('message_digest', message_digest),
                    simple_cms_attribute(
                        'cms_algorithm_protection', algo_protection
                    ),
                ]
            )

    return Handler


@pdfmac_with_handler("MAC mechanism does not match")
def test_cms_algorithm_protection_mac_algo_mismatch():
    class Handler(pdfmac.PdfMacTokenHandler):
        def _format_auth_attrs(
            self, message_digest: bytes
        ) -> cms.CMSAttributes:
            algo_protection = cms.CMSAlgorithmProtection(
                {
                    'mac_algorithm': cms.HmacAlgorithm({'algorithm': 'sha512'}),
                    'digest_algorithm': cms.DigestAlgorithm(
                        {'algorithm': self.md_algorithm}
                    ),
                }
            )
            return cms.CMSAttributes(
                [
                    simple_cms_attribute(
                        'content_type', 'pdf_mac_integrity_info'
                    ),
                    simple_cms_attribute('message_digest', message_digest),
                    simple_cms_attribute(
                        'cms_algorithm_protection', algo_protection
                    ),
                ]
            )

    return Handler


@pdfmac_with_handler("attribute not valid for authenticated data")
def test_cms_algorithm_protection_mac_algo_missing():
    class Handler(pdfmac.PdfMacTokenHandler):
        def _format_auth_attrs(
            self, message_digest: bytes
        ) -> cms.CMSAttributes:
            algo_protection = cms.CMSAlgorithmProtection(
                {
                    'digest_algorithm': cms.DigestAlgorithm(
                        {'algorithm': self.md_algorithm}
                    ),
                }
            )
            return cms.CMSAttributes(
                [
                    simple_cms_attribute(
                        'content_type', 'pdf_mac_integrity_info'
                    ),
                    simple_cms_attribute('message_digest', message_digest),
                    simple_cms_attribute(
                        'cms_algorithm_protection', algo_protection
                    ),
                ]
            )

    return Handler


@pdfmac_with_handler("Content type not found in authenticated attributes")
def test_duplicate_content_type_attr():
    class Handler(pdfmac.PdfMacTokenHandler):
        def _format_auth_attrs(
            self, message_digest: bytes
        ) -> cms.CMSAttributes:
            return cms.CMSAttributes(
                [
                    cms.CMSAttribute(
                        {
                            'type': cms.CMSAttributeType('content_type'),
                            'values': ('data', 'pdf_mac_integrity_info'),
                        }
                    ),
                    simple_cms_attribute('message_digest', message_digest),
                ]
            )

    return Handler


@pdfmac_with_handler("Content type not found in authenticated attributes")
def test_no_content_type_attr():
    class Handler(pdfmac.PdfMacTokenHandler):
        def _format_auth_attrs(
            self, message_digest: bytes
        ) -> cms.CMSAttributes:
            return cms.CMSAttributes(
                [
                    simple_cms_attribute('message_digest', message_digest),
                ]
            )

    return Handler


@pdfmac_with_handler("content type.*must be id-pdfMacIntegrityInfo")
def test_wrong_content_type_attr():
    class Handler(pdfmac.PdfMacTokenHandler):
        def _format_auth_attrs(
            self, message_digest: bytes
        ) -> cms.CMSAttributes:
            return cms.CMSAttributes(
                [
                    simple_cms_attribute('content_type', 'data'),
                    simple_cms_attribute('message_digest', message_digest),
                ]
            )

    return Handler


@pdfmac_with_handler("Message digest not found")
def test_no_md():
    class Handler(pdfmac.PdfMacTokenHandler):
        def _format_auth_attrs(
            self, message_digest: bytes
        ) -> cms.CMSAttributes:
            return cms.CMSAttributes(
                [
                    simple_cms_attribute(
                        'content_type', 'pdf_mac_integrity_info'
                    ),
                ]
            )

    return Handler


@pdfmac_with_handler("messageDigest attribute does not match")
def test_bad_md():
    class Handler(pdfmac.PdfMacTokenHandler):
        def _format_auth_attrs(
            self, message_digest: bytes
        ) -> cms.CMSAttributes:
            dummy = hashlib.sha256(b'\xde\xad\xbe\xef').digest()
            return cms.CMSAttributes(
                [
                    simple_cms_attribute(
                        'content_type', 'pdf_mac_integrity_info'
                    ),
                    simple_cms_attribute('message_digest', dummy),
                ]
            )

    return Handler


@pdfmac_with_handler("PDF MAC tokens cannot have unauthenticated attributes")
def test_with_unauth_attrs():
    class Handler(pdfmac.PdfMacTokenHandler):
        def build_pdfmac_token(self, *args, **kwargs) -> cms.ContentInfo:
            token = super().build_pdfmac_token(*args, **kwargs)
            # doesn't have to make any sense
            token['content']['unauth_attrs'] = cms.CMSAttributes(
                [simple_cms_attribute('content_type', 'data')]
            )
            return token

    return Handler


@pdfmac_with_handler("key derivation.*identified as pdfMacWrapKdf")
def test_no_kdf():
    class Handler(pdfmac.PdfMacTokenHandler):
        def _get_mac_keying_info(self, dry_run):
            mac_key, ri = super()._get_mac_keying_info(dry_run)
            del ri.chosen['key_derivation_algorithm']
            return mac_key, ri

    return Handler


@pdfmac_with_handler("requires exactly one recipientInfo")
def test_too_many_recipients():
    class Handler(pdfmac.PdfMacTokenHandler):
        def _format_auth_data(self, *args, **kwargs) -> cms.AuthenticatedData:
            ad = super()._format_auth_data(*args, **kwargs)
            ris = ad['recipient_infos']
            ris.append(ris[0])
            return ad

    return Handler


@pdfmac_with_handler("must be.*PasswordRecipientInfo")
def test_wrong_recipient_type():
    class Handler(pdfmac.PdfMacTokenHandler):
        def _format_auth_data(
            self, recipient_info, *args, **kwargs
        ) -> cms.AuthenticatedData:
            return super()._format_auth_data(
                recipient_info=cms.RecipientInfo(
                    {
                        'ori': cms.OtherRecipientInfo(
                            {
                                'ori_type': '2.999',
                                'ori_value': core.OctetString(
                                    b'\xde\xad\xbe\xef'
                                ),
                            }
                        )
                    }
                ),
                *args,
                **kwargs,
            )

    return Handler


@pdfmac_with_handler("must be.*pdfMacIntegrityInfo")
def test_wrong_enc_content_type():
    class Handler(pdfmac.PdfMacTokenHandler):
        def _format_auth_attrs(
            self, message_digest: bytes
        ) -> cms.CMSAttributes:
            # the signers we currently use in the tests will use SHA-384
            hash_fun = getattr(hashlib, self.md_algorithm)
            return super()._format_auth_attrs(
                hash_fun(b'\xde\xad\xbe\xef').digest()
            )

        def _format_auth_data(self, *args, **kwargs) -> cms.AuthenticatedData:
            ad = super()._format_auth_data(*args, **kwargs)
            ad['encap_content_info'] = cms.EncapsulatedContentInfo(
                {
                    'content_type': 'data',
                    'content': core.ParsableOctetString(b'\xde\xad\xbe\xef'),
                }
            )
            return ad

    return Handler


@pdfmac_with_handler("must be.*type AuthenticatedData")
def test_wrong_token_content_type():
    class Handler(pdfmac.PdfMacTokenHandler):
        def build_pdfmac_token(self, *args, **kwargs) -> cms.ContentInfo:
            return cms.ContentInfo(
                {
                    'content_type': 'data',
                    'content': core.OctetString(b'\xde\xad\xbe\xef'),
                }
            )

    return Handler


@pdfmac_with_handler(
    "unexpected signature digest", only_location=MacLocation.STANDALONE
)
def test_unexpected_sig_digest():
    class Handler(pdfmac.PdfMacTokenHandler):
        def _format_message(self, document_digest, signature_digest):
            return super()._format_message(document_digest, document_digest)

    return Handler


@pdfmac_with_handler(
    "could not find signature digest", only_location=MacLocation.IN_SIG
)
def test_missing_sig_digest():
    class Handler(pdfmac.PdfMacTokenHandler):
        def _format_message(self, document_digest, signature_digest):
            return super()._format_message(
                document_digest, signature_digest=None
            )

    return Handler


@pdfmac_with_handler(
    "Signature digest does not match value in", only_location=MacLocation.IN_SIG
)
def test_sig_wrong_sig_digest():
    class Handler(pdfmac.PdfMacTokenHandler):
        def _format_message(self, document_digest, signature_digest):
            dummy = hashlib.sha256(b'\xde\xad\xbe\xef').digest()
            return super()._format_message(
                document_digest, signature_digest=dummy
            )

    return Handler


@pdfmac_with_handler(
    "Document digest does not match value in", only_location=MacLocation.IN_SIG
)
def test_sig_wrong_document_digest():
    class Handler(pdfmac.PdfMacTokenHandler):
        def _format_message(self, document_digest, signature_digest):
            dummy = hashlib.sha256(b'\xde\xad\xbe\xef').digest()
            return super()._format_message(
                dummy, signature_digest=signature_digest
            )

    return Handler


@pdfmac_with_handler(
    "must not have trailing CMS data", only_location=MacLocation.STANDALONE
)
def test_with_trailing_data():
    class Handler(pdfmac.PdfMacTokenHandler):
        def build_pdfmac_token(self, *args, dry_run: bool = False, **kwargs):
            token = super().build_pdfmac_token(*args, **kwargs, dry_run=dry_run)
            if dry_run:
                # just waste some space, this is only used for the estimate
                # anyhow, and all we need is an estimate that's a bit too large
                token['content']['unauth_attrs'] = cms.CMSAttributes(
                    [simple_cms_attribute('content_type', 'data')]
                )
            return token

    return Handler


@pdfmac_known_bad_case("Error retrieving salt")
@pytest.mark.parametrize('encryption_type', list(EncryptionType))
def test_do_not_store_salt(encryption_type):
    w = init_sample_doc(encryption_type)
    del w._encrypt.get_object()['/KDFSalt']
    out = BytesIO()
    w.write(out)
    return out


@pdfmac_known_bad_case("must have /ByteRange covering the entire file")
@pytest.mark.parametrize('encryption_type', list(EncryptionType))
def test_mac_no_full_coverage(encryption_type):
    w = init_sample_doc(encryption_type)
    out = BytesIO()
    w.write(out)

    r = PdfFileReader(out)
    _dummy_decrypt(r, ignore_mac=True)
    w = IncrementalPdfFileWriter.from_reader(r)
    # short-circuit MAC
    w.digest_aware_write = True
    w.root['/Foo'] = generic.pdf_name('/Bar')
    w.update_root()
    w.write_in_place()
    return out


@pdfmac_known_bad_case("must have /ByteRange covering the entire file")
@pytest.mark.parametrize('encryption_type', list(EncryptionType))
def test_mac_with_signature_no_full_coverage(encryption_type):
    w = init_sample_doc(encryption_type)
    signer = SimpleSigner(
        signing_cert=FROM_ECC_CA.signing_cert,
        signing_key=FROM_ECC_CA.signing_key,
        cert_registry=FROM_ECC_CA.cert_registry,
    )
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(field_name='Sig'),
        signer=signer,
    )

    r = PdfFileReader(out)
    _dummy_decrypt(r, ignore_mac=True)
    w = IncrementalPdfFileWriter.from_reader(r)
    # short-circuit MAC
    w.digest_aware_write = True
    w.root['/Foo'] = generic.pdf_name('/Bar')
    w.update_root()
    w.write_in_place()
    return out


def test_simple_mac_weak_hash():
    w = init_sample_doc(EncryptionType.STANDARD)

    class Handler(pdfmac.PdfMacTokenHandler):
        @classmethod
        def from_key_mat(
            cls, file_encryption_key: bytes, kdf_salt: bytes, md_algorithm: str
        ):
            mac_kek = cls._derive_mac_kek(file_encryption_key, kdf_salt)
            return cls(mac_kek=mac_kek, md_algorithm='sha1')

    w._mac_handler_cls = Handler
    out = BytesIO()
    w.write(out)
    r = PdfFileReader(out)
    with pytest.raises(DisallowedAlgorithmError):
        _dummy_decrypt(r)


@pdfmac_known_bad_case("Failed to locate AuthCode dictionary")
@pytest.mark.parametrize(
    'encryption_type,remove_salt',
    list(itertools.product(EncryptionType, [True, False])),
)
def test_stripped_mac_detected(encryption_type, remove_salt):
    class Writer(PdfFileWriter):
        def write(self, stream):
            # skip writing the MAC
            self._write(stream)

    w = init_sample_doc(encryption_type, writer_class=Writer)
    if remove_salt:
        del w._encrypt.get_object()['/KDFSalt']
    out = BytesIO()
    w.write(out)
    return out


@pdfmac_known_bad_case("AuthCode.*indirect")
@pytest.mark.parametrize(
    'encryption_type,mac_location',
    list(itertools.product(EncryptionType, MacLocation)),
)
def test_ac_dict_indirect(encryption_type, mac_location):
    class Writer(PdfFileWriter):
        def set_custom_trailer_entry(
            self, key: generic.NameObject, value: generic.PdfObject
        ):
            if key == '/AuthCode':
                value = self.add_object(value)
            super().set_custom_trailer_entry(key, value)

    w = init_sample_doc(encryption_type, writer_class=Writer)
    if mac_location == MacLocation.IN_SIG:
        signer = SimpleSigner(
            signing_cert=FROM_ECC_CA.signing_cert,
            signing_key=FROM_ECC_CA.signing_key,
            cert_registry=FROM_ECC_CA.cert_registry,
        )
        return signers.sign_pdf(
            w,
            signers.PdfSignatureMetadata(field_name='Sig'),
            signer=signer,
        )
    else:
        out = BytesIO()
        w.write(out)
        return out


def _manipulate_with_sig(encryption_type, manipulator):
    class Writer(PdfFileWriter):
        def set_custom_trailer_entry(
            self, key: generic.NameObject, value: generic.PdfObject
        ):
            if key == '/AuthCode':
                assert isinstance(value, generic.DictionaryObject)
                manipulator(self, value)
            super().set_custom_trailer_entry(key, value)

    w = init_sample_doc(encryption_type, writer_class=Writer)

    signer = SimpleSigner(
        signing_cert=FROM_ECC_CA.signing_cert,
        signing_key=FROM_ECC_CA.signing_key,
        cert_registry=FROM_ECC_CA.cert_registry,
    )
    return signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(field_name='Sig'),
        signer=signer,
    )


def _manipulate_standalone(encryption_type, manipulator):
    class Writer(PdfFileWriter):
        def set_custom_trailer_entry(
            self, key: generic.NameObject, value: generic.PdfObject
        ):
            if key == '/AuthCode':
                assert isinstance(value, generic.DictionaryObject)
                manipulator(self, value)
            super().set_custom_trailer_entry(key, value)

    w = init_sample_doc(encryption_type, writer_class=Writer)
    out = BytesIO()
    w.write(out)
    return out


@pdfmac_known_bad_case("requires /SigObjRef")
@pytest.mark.parametrize('encryption_type', list(EncryptionType))
def test_missing_sig_ref(encryption_type):
    def manipulate(_w, value):
        del value['/SigObjRef']

    return _manipulate_with_sig(encryption_type, manipulator=manipulate)


@pdfmac_known_bad_case("must be an indirect reference")
@pytest.mark.parametrize('encryption_type', list(EncryptionType))
def test_malformed_sig_ref_direct(encryption_type):
    def manipulate(_w, value):
        value['/SigObjRef'] = generic.DictionaryObject()

    return _manipulate_with_sig(encryption_type, manipulator=manipulate)


@pdfmac_known_bad_case("does not point to a dictionary")
@pytest.mark.parametrize('encryption_type', list(EncryptionType))
def test_malformed_sig_ref_not_dictionary(encryption_type):
    def manipulate(w, value):
        value['/SigObjRef'] = w.add_object(generic.pdf_name('/Bleh'))

    return _manipulate_with_sig(encryption_type, manipulator=manipulate)


@pdfmac_known_bad_case("Failed to retrieve signature contents")
@pytest.mark.parametrize('encryption_type', list(EncryptionType))
def test_malformed_sig_ref_not_signature(encryption_type):
    def manipulate(w, value):
        value['/SigObjRef'] = w.add_object(generic.DictionaryObject())

    return _manipulate_with_sig(encryption_type, manipulator=manipulate)


@pdfmac_known_bad_case("Failed to locate MAC in document")
def test_mac_location_missing():
    def manipulate(_w, value):
        del value['/MACLocation']

    return _manipulate_standalone(
        EncryptionType.STANDARD, manipulator=manipulate
    )


@pdfmac_known_bad_case("Failed to locate MAC in document")
def test_mac_location_unsupported():
    def manipulate(_w, value):
        value['/MACLocation'] = generic.pdf_name('/Bleh')

    return _manipulate_standalone(
        EncryptionType.STANDARD, manipulator=manipulate
    )


@pdfmac_known_bad_case("Failed to retrieve standalone MAC value")
@pytest.mark.parametrize('encryption_type', list(EncryptionType))
def test_mac_wrong_type(encryption_type):
    region_start_lazy: Optional[Callable[[], int]] = None

    def manipulate(_w, value):
        nonlocal region_start_lazy
        # we have to get a little creative here, since overriding the /MAC
        # field directly will break the serialisation logic
        region_start_lazy = lambda: value['/ByteRange'].first_region_len

    out = _manipulate_standalone(encryption_type, manipulator=manipulate)

    out.seek(region_start_lazy())
    out.write(b'/Bleh%')

    return out


@pdfmac_known_bad_case("No sensible /ByteRange found")
def test_byte_range_missing():
    # again, we need some creativity here
    br_obj = None

    def manipulate(_w, value):
        nonlocal br_obj
        br_obj = value['/ByteRange']

    out = _manipulate_standalone(
        EncryptionType.STANDARD, manipulator=manipulate
    )
    assert isinstance(br_obj, SigByteRangeObject)
    out.seek(br_obj._range_object_offset - 3)
    out.write(b'%\n/Bleh%')
    return out


def manipulatable_sign(manipulate_cms):
    # CMS-agnostic signing example
    #
    # write an in-place certification signature using the PdfCMSEmbedder
    # low-level API directly.

    w = init_sample_doc(EncryptionType.STANDARD)

    cms_writer = cms_embedder.PdfCMSEmbedder().write_cms(
        field_name='Signature', writer=w
    )
    next(cms_writer)

    sig_obj = signers.SignatureObject(bytes_reserved=8192)

    md_algorithm = 'sha256'
    sig_obj_ref = cms_writer.send(
        cms_embedder.SigObjSetup(
            sig_placeholder=sig_obj,
        )
    )

    ac_dict = generic.DictionaryObject(
        {
            generic.pdf_name('/MACLocation'): generic.pdf_name(
                '/AttachedToSig'
            ),
            generic.pdf_name('/SigObjRef'): sig_obj_ref,
        }
    )
    w.set_custom_trailer_entry(generic.pdf_name('/AuthCode'), ac_dict)

    prep_digest, output = cms_writer.send(
        cms_embedder.SigIOSetup(md_algorithm=md_algorithm)
    )

    signer: signers.SimpleSigner = FROM_ECC_CA
    content_info = asyncio.run(
        signer.async_sign(
            data_digest=prep_digest.document_digest,
            digest_algorithm=md_algorithm,
        )
    )
    manipulate_cms(content_info)
    cms_writer.send(content_info)

    return output


@pdfmac_known_bad_case("exactly 1 signerInfo")
def test_signer_info_cardinality_multi():
    def manipulate(ci):
        signer_infos = ci['content']['signer_infos']
        signer_infos.append(signer_infos[0])

    return manipulatable_sign(manipulate)


@pdfmac_known_bad_case("exactly 1 signerInfo")
def test_signer_info_cardinality_zero():
    def manipulate(ci):
        ci['content']['signer_infos'] = cms.SignerInfos([])

    return manipulatable_sign(manipulate)


@pdfmac_known_bad_case("exactly 1 pdfMacData unsigned attribute")
def test_pdf_mac_attr_missing():
    return manipulatable_sign(id)


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_sign_cannot_reuse_pdf_signer_with_macs_in_parallel():
    w1 = init_sample_doc(EncryptionType.STANDARD)
    w2 = init_sample_doc(EncryptionType.STANDARD)
    signer = SimpleSigner(
        signing_cert=FROM_ECC_CA.signing_cert,
        signing_key=FROM_ECC_CA.signing_key,
        cert_registry=FROM_ECC_CA.cert_registry,
    )
    pdf_signer = PdfSigner(
        signers.PdfSignatureMetadata(field_name='Sig'),
        signer=signer,
        timestamper=DUMMY_TS,
    )

    with pytest.raises(
        SigningError, match='Other PDF MAC attribute provider found'
    ):
        await asyncio.gather(
            pdf_signer.async_sign_pdf(w1), pdf_signer.async_sign_pdf(w2)
        )
